"""Support Vulture."""
from argparse import ArgumentParser

from vulture.core import ERROR_CODES, Vulture, make_config

from pylama.context import RunContext
from pylama.lint import LinterV2 as BaseLinter


class Linter(BaseLinter):
    """vulture runner."""

    name = "vulture"

    @classmethod
    def add_args(cls, parser: ArgumentParser):
        """Add --max-complexity option."""
        parser.add_argument(
            "--vulture-min-confidence",
            type=int,
            help="Minimum confidence (between 0 and 100) for code to be reported as unused.",
        )
        parser.add_argument(
            "--vulture-ignore-names",
            help="Comma-separated list of names to ignore",
        )
        parser.add_argument(
            "--vulture-ignore-decorators",
            help="Comma-separated list of decorators to ignore",
        )

    def run_check(self, ctx: RunContext):  # noqa
        """Check code with vulture."""
        params = ctx.get_params("vulture")
        options = ctx.options
        if options:
            params.setdefault("min-confidence", options.vulture_min_confidence)
            params.setdefault("ignore-names", options.vulture_ignore_names)
            params.setdefault("ignore-decorators", options.vulture_ignore_decorators)

        config = make_config(parse_params(ctx.filename, params))
        vulture = Vulture(
            verbose=config["verbose"],
            ignore_names=config["ignore_names"],
            ignore_decorators=config["ignore_decorators"],
        )
        vulture.scan(ctx.source, filename=ctx.filename)
        unused_code_items = vulture.get_unused_code(
            min_confidence=config["min_confidence"], sort_by_size=config["sort_by_size"]
        )
        for item in unused_code_items:
            error_code = ERROR_CODES[item.typ]
            ctx.push(
                source="vulture",
                type="R",
                lnum=item.first_lineno,
                number=error_code,
                text=f"{item.message} ({item.confidence}% confidence)",
            )


def parse_params(path, params=None):
    """Convert params from pylama."""
    return [f"--{key}={value}" for key, value in params.items() if value] + [path]



"""Pylama TOML configuration."""

import toml

from pylama.libs.inirama import Namespace as _Namespace


class Namespace(_Namespace):
    """Inirama-style wrapper for TOML config."""

    def parse(self, source: str, update: bool = True, **params):
        """Parse TOML source as string."""
        content = toml.loads(source)
        tool = content.get("tool", {})
        pylama = tool.get("pylama", {})
        linters = pylama.pop("linter", {})
        files = pylama.pop("files", [])

        for name, value in pylama.items():
            self["pylama"][name] = value

        for linter, options in linters.items():
            for name, value in options.items():
                self[f"pylama:{linter}"][name] = value

        for file in files:
            path = file.pop("path", None)
            if path is None:
                continue
            for name, value in file.items():
                self[f"pylama:{path}"][name] = value



"""Support Radon.

Supports stdin.
"""

from argparse import ArgumentError, ArgumentParser

from radon.complexity import add_inner_blocks
from radon.visitors import ComplexityVisitor

from pylama.context import RunContext
from pylama.lint import LinterV2 as Abstract


class Linter(Abstract):
    """Radon runner."""

    name = "radon"

    @classmethod
    def add_args(cls, parser: ArgumentParser):
        """Add --max-complexity option."""
        parser.add_argument(
            "--radon-no-assert",
            default=False,
            action="store_true",
            help="Ignore `assert` statements.",
        )
        parser.add_argument(
            "--radon-show-closures",
            default=False,
            action="store_true",
            help="Increase complexity on closures.",
        )
        try:
            parser.add_argument(
                "--max-complexity",
                default=10,
                type=int,
                help="Max complexity threshold",
            )
        except ArgumentError:
            pass

    def run_check(self, ctx: RunContext):  # noqa  # noqa
        """Check code with Radon."""
        params = ctx.get_params("radon")
        options = ctx.options
        if options:
            params.setdefault("complexity", options.max_complexity)
            params.setdefault("no_assert", options.radon_no_assert)
            params.setdefault("show_closures", options.radon_show_closures)

        complexity = params.get("complexity", 10)
        no_assert = params.get("no_assert", False)
        show_closures = params.get("show_closures", False)
        visitor = ComplexityVisitor.from_code(ctx.source, no_assert=no_assert)
        blocks = visitor.blocks
        if show_closures:
            blocks = add_inner_blocks(blocks)
        for block in visitor.blocks:
            if block.complexity > complexity:
                ctx.push(
                    lnum=block.lineno,
                    col=block.col_offset + 1,
                    source="radon",
                    type="R",
                    number="R901",
                    text=f"{block.name} is too complex {block.complexity}",
                )



"""Commented-out code checking."""

from eradicate import Eradicator

from pylama.context import RunContext
from pylama.lint import LinterV2 as Abstract


class Linter(Abstract):
    """Run commented-out code checking."""

    name = "eradicate"

    def run_check(self, ctx: RunContext):
        """Eradicate code checking.

        TODO: Support params
        """
        eradicator = Eradicator()
        line_numbers = eradicator.commented_out_code_line_numbers(ctx.source)
        for line_number in line_numbers:
            ctx.push(
                lnum=line_number,
                source="eradicate",
                text=str("Found commented out code"),
                number="E800",
                type="E",
            )



""" py.test plugin for checking files with pylama. """
from __future__ import absolute_import

import pathlib
from os import path as op

import pytest

from pylama.config import CURDIR
from pylama.main import DEFAULT_FORMAT, check_paths, parse_options

HISTKEY = "pylama/mtimes"


def pytest_load_initial_conftests(early_config, *_):
    # Marks have to be registered before usage
    # to not fail with --strict command line argument
    early_config.addinivalue_line(
        "markers", "pycodestyle: Mark test as using pylama code audit tool."
    )


def pytest_addoption(parser):
    group = parser.getgroup("general")
    group.addoption(
        "--pylama",
        action="store_true",
        help="perform some pylama code checks on .py files",
    )


def pytest_sessionstart(session):
    config = session.config
    if config.option.pylama and getattr(config, "cache", None):
        config._pylamamtimes = config.cache.get(HISTKEY, {})


def pytest_sessionfinish(session):
    config = session.config
    if hasattr(config, "_pylamamtimes"):
        config.cache.set(HISTKEY, config._pylamamtimes)


def pytest_collect_file(path, parent):
    config = parent.config
    if config.option.pylama and path.ext == ".py":
        return PylamaFile.from_parent(parent, path=pathlib.Path(path))
    return None


class PylamaError(Exception):
    """indicates an error during pylama checks."""


class PylamaFile(pytest.File):
    def collect(self):
        return [PylamaItem.from_parent(self, name="pylama")]


class PylamaItem(pytest.Item):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_marker("pycodestyle")
        self.cache = None
        self._pylamamtimes = None

    def setup(self):
        if not getattr(self.config, "cache", None):
            return False

        self.cache = True
        self._pylamamtimes = self.fspath.mtime()
        pylamamtimes = self.config._pylamamtimes
        old = pylamamtimes.get(str(self.fspath), 0)
        if old == self._pylamamtimes:
            pytest.skip("file(s) previously passed Pylama checks")

        return True

    def runtest(self):
        errors = check_file(self.fspath)
        if errors:
            out = "\n".join(err.format(DEFAULT_FORMAT) for err in errors)
            raise PylamaError(out)

        # update mtime only if test passed
        # otherwise failures would not be re-run next time
        if self.cache:
            self.config._pylamamtimes[str(self.fspath)] = self._pylamamtimes

    def repr_failure(self, excinfo, style=None):
        if excinfo.errisinstance(PylamaError):
            return excinfo.value.args[0]
        return super().repr_failure(excinfo, style)


def check_file(path):
    options = parse_options()
    path = op.relpath(str(path), CURDIR)
    return check_paths([path], options, rootdir=CURDIR)


# pylama:ignore=D,E1002,W0212,F0001,C0115,C0116



""" Don't duplicate same errors from different linters. """
from __future__ import annotations

import re
from collections import defaultdict
from typing import Any, DefaultDict, Dict, Generator, List, Set, Tuple

PATTERN_NUMBER = re.compile(r"^\s*([A-Z]\d+)\s*", re.I)

DUPLICATES: Dict[Tuple[str, str], Set] = {
    key: values  # type: ignore
    for values in (
        # multiple statements on one line
        {("pycodestyle", "E701"), ("pylint", "C0321")},
        # unused variable
        {("pylint", "W0612"), ("pyflakes", "W0612")},
        # undefined variable
        {("pylint", "E0602"), ("pyflakes", "E0602")},
        # unused import
        {("pylint", "W0611"), ("pyflakes", "W0611")},
        # whitespace before ')'
        {("pylint", "C0326"), ("pycodestyle", "E202")},
        # whitespace before '('
        {("pylint", "C0326"), ("pycodestyle", "E211")},
        # multiple spaces after operator
        {("pylint", "C0326"), ("pycodestyle", "E222")},
        # missing whitespace around operator
        {("pylint", "C0326"), ("pycodestyle", "E225")},
        # unexpected spaces
        {("pylint", "C0326"), ("pycodestyle", "E251")},
        # long lines
        {("pylint", "C0301"), ("pycodestyle", "E501")},
        # statement ends with a semicolon
        {("pylint", "W0301"), ("pycodestyle", "E703")},
        # multiple statements on one line
        {('pylint", "C0321'), ("pycodestyle", "E702")},
        # bad indentation
        {("pylint", "W0311"), ("pycodestyle", "E111")},
        # wildcart import
        {("pylint", "W00401"), ("pyflakes", "W0401")},
        # module docstring
        {("pydocstyle", "D100"), ("pylint", "C0111")},
    )
    for key in values  # type: ignore
}


class Error:
    """Store an error's information."""

    __slots__ = "source", "col", "lnum", "etype", "message", "filename", "number"

    def __init__(
        self,
        source="pylama",
        col=1,
        lnum=1,
        type=None,  # pylint: disable=R0913
        text="unknown error",
        filename="",
        number="",
        **_,
    ):
        """Init error information with default values."""
        text = str(text).strip().replace("\n", " ")
        if number:
            self.number = number
        else:
            number = PATTERN_NUMBER.match(text)
            self.number = number.group(1).upper() if number else ""

        self.etype = type[:1] if type else (number[0] if number else "E")
        self.col = max(col, 1)
        self.filename = filename
        self.source = source
        self.lnum = int(lnum)
        self.message = text

    def __repr__(self):
        return f"<Error:{self.lnum}:{self.col}: {self.number} {self.message}>"

    def format(self, pattern: str) -> str:
        """Format the error with the given pattern."""
        return pattern.format(
            filename=self.filename,
            lnum=self.lnum,
            col=self.col,
            message=self.message,
            etype=self.etype,
            source=self.source,
            number=self.number,
        )

    def to_dict(self) -> Dict[str, Any]:
        """Return the error as a dict."""
        return {
            "source": self.source,
            "col": self.col,
            "lnum": self.lnum,
            "etype": self.etype,
            "message": self.message,
            "filename": self.filename,
            "number": self.number,
        }


def remove_duplicates(errors: List[Error]) -> Generator[Error, None, None]:
    """Filter duplicates from given error's list."""
    passed: DefaultDict[int, Set] = defaultdict(set)
    for error in errors:
        key = error.source, error.number
        if key in DUPLICATES:
            if key in passed[error.lnum]:
                continue
            passed[error.lnum] = DUPLICATES[key]
        yield error


def default_sorter(err: Error) -> Any:
    """Sort by line number."""
    return err.lnum


# pylama:ignore=W0622,D,R0924



"""Manage resources."""

import ast
import os.path as op
import re
from argparse import Namespace
from copy import copy
from functools import lru_cache
from pathlib import Path
from tempfile import NamedTemporaryFile, mkdtemp
from typing import Dict, List, Set

from pylama.errors import Error
from pylama.utils import read

# Parse modeline
MODELINE_RE = re.compile(
    r"^\s*#\s+(?:pylama:)\s*((?:[\w_]*=[^:\n\s]+:?)+)", re.I | re.M
).search

SKIP_PATTERN = re.compile(r"# *noqa\b", re.I).search


class RunContext:  # pylint: disable=R0902
    """Manage resources."""

    __slots__ = (
        "errors",
        "options",
        "skip",
        "ignore",
        "select",
        "linters",
        "linters_params",
        "filename",
        "_ast",
        "_from_stdin",
        "_source",
        "_tempfile",
        "_lines",
    )

    def __init__(self, filename: str, source: str = None, options: Namespace = None):
        """Initialize the class."""
        self.errors: List[Error] = []
        self.options = options
        self.skip = False
        self.ignore = set()
        self.select = set()
        self.linters = []
        self.linters_params = {}

        self._ast = None
        self._from_stdin = source is not None
        self._source = source
        self._tempfile = None
        self._lines = None

        if options:
            if options.abspath:
                filename = op.abspath(filename)
            self.skip = options.skip and any(
                ptrn.match(filename) for ptrn in options.skip
            )
            self.linters.extend(options.linters)
            self.ignore |= options.ignore
            self.select |= options.select
            self.linters_params.update(options.linters_params)

            for mask in options.file_params:
                if mask.match(filename):
                    fparams = options.file_params[mask]
                    self.update_params(**fparams)

        self.filename = filename

        # Read/parse modeline
        if not self.skip:
            modeline = MODELINE_RE(self.source)
            if modeline:
                values = modeline.group(1).split(":")
                self.update_params(**dict(v.split("=", 1) for v in values))  # type: ignore

    def __enter__(self):
        """Enter to context."""
        return self

    def __exit__(self, etype, evalue, _):
        """Exit from the context."""
        if self._tempfile is not None:
            tmpfile = Path(self._tempfile)
            tmpfile.unlink()
            tmpfile.parent.rmdir()

        if evalue is not None:
            if etype is IOError:
                self.push(text=f"{evalue}", number="E001")
            elif etype is UnicodeDecodeError:
                self.push(text=f"UnicodeError: {self.filename}", number="E001")
            elif etype is SyntaxError:
                self.push(
                    lnum=evalue.lineno,
                    col=evalue.offset,
                    text=f"SyntaxError: {evalue.args[0]}",
                )
            else:
                self.push(lnum=1, col=1, text=str(evalue))
                return False

        return True

    @property
    def source(self):
        """Get the current source code."""
        if self._source is None:
            self._source = read(self.filename)
        return self._source

    @property
    def lines(self):
        """Split source to lines."""
        if self._lines is None:
            self._lines = self.source.splitlines(True)
        return self._lines

    @property
    def ast(self):
        """Get the AST for the source."""
        if self._ast is None:
            self._ast = compile(self.source, self.filename, "exec", ast.PyCF_ONLY_AST)
        return self._ast

    @property
    def temp_filename(self):
        """Get a filename for run external command."""
        if not self._from_stdin:
            return self.filename

        if self._tempfile is None:
            file = NamedTemporaryFile(  # noqa
                "w",
                encoding="utf8",
                suffix=".py",
                dir=mkdtemp(prefix="pylama_"),
                delete=False,
            )
            file.write(self.source)
            file.close()
            self._tempfile = file.name

        return self._tempfile

    def update_params(self, ignore=None, select=None, linters=None, skip=None, **_):
        """Update general params (from file configs or modeline)."""
        if select:
            self.select |= set(select.split(","))
        if ignore:
            self.ignore |= set(ignore.split(","))
        if linters:
            self.linters = linters.split(",")
        if skip is not None:
            self.skip = bool(int(skip))

    @lru_cache(42)
    def get_params(self, name: str) -> Dict:
        """Get params for a linter with the given name."""
        lparams = copy(self.linters_params.get(name, {}))
        for key in ("ignore", "select"):
            if key in lparams and not isinstance(lparams[key], set):
                lparams[key] = set(lparams[key].split(","))
        return lparams

    @lru_cache(42)
    def get_filter(self, name: str, key: str) -> Set:
        """Get select/ignore from linter params."""
        lparams = self.get_params(name)
        return lparams.get(key, set())

    def push(self, filtrate: bool = True, **params):
        """Record an error."""
        err = Error(filename=self.filename, **params)
        number = err.number

        if len(self.lines) >= err.lnum and SKIP_PATTERN(self.lines[err.lnum - 1]):
            return None

        if filtrate:

            for rule in self.select | self.get_filter(err.source, "select"):
                if number.startswith(rule):
                    return self.errors.append(err)

            for rule in self.ignore | self.get_filter(err.source, "ignore"):
                if number.startswith(rule):
                    return None

        return self.errors.append(err)
